@@ -31,6 +31,9 @@
  * be.
  */
 
+#include "basename.h"
+/* for curlx_basename() function */
+
 #include "binmode.h"
 /* "binmode.h" provides macro CURLX_SET_BINMODE() */
 
@@ -55,6 +58,9 @@
 #include "strparse.h"
 /* The curlx_str_* parsing functions */
 
+#include "strcopy.h"
+/* curlx_strcopy */
+
 #include "dynbuf.h"
 /* The curlx_dyn_* functions */
 
